/**@file
 * AskSkipOverwriteDialog.m
 * @brief Ask skip or overwrite dialog box class.
 * @author toru kageyama
 * Created by Toru Kageyama on 10/10/12.
 * Copyright 2010 Comona Co., Ltd. All rights reserved.
 */

#import "AskSkipOverwriteDialog.h"
#import "MyDocument.h"
#import "ViewController.h"

@implementation AskSkipOverwriteDialog

@synthesize m_filePath;

/**
 * Base class retrieves sheet pointer.
 * @return sheet ID.
 */
- (id)sheet
{
	[m_doSame setState:NSOffState];
	[m_title setStringValue:[NSString stringWithFormat:NSLocalizedString(@"same path exist", @"same path exist"), m_filePath]];
	id controller = [m_document controller];
	return [controller askSkipOverwriteDialog];
}

/**
 * skip button pressed.
 * @param[in] sender the sender object.
 */
- (IBAction)skipButton:(id)sender
{
	// skip button is pressed.
	id controller = [m_document controller];
	[controller replaySkip:YES doSame:[m_doSame state] == NSOnState];
	[m_sheet close];
	[NSApp stopModal];
}

/**
 * overwrite button pressed.
 * @param[in] sender the sender object.
 */
- (IBAction)overwriteButton:(id)sender
{
	// overwrite button is pressed.
	id controller = [m_document controller];
	[controller replaySkip:NO doSame:[m_doSame state] == NSOnState];
	[m_sheet close];
	[NSApp stopModal];
}

// show modal.
- (void)showModal
{
	m_sheet = [self sheet];
	[NSApp runModalForWindow:m_sheet];
}

// destructor.
- (void)dealloc
{
	[m_filePath release];
	[super dealloc];
}

@end
