/**@file
 * Base64.m
 * @brief Base64 encode/decode class.
 * @author toru kageyama
 * Created by Toru Kageyama on 10/09/15.
 * Copyright 2010 Comona Co., Ltd. All rights reserved.
 */

#import "Base64.h"

// constant.
static const char base64Str[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
static const char padChar = '=';
static const NSUInteger bit_sfr[] = { 26, 20, 14, 8 };
//
// Mapping from ASCII character to 6 bit pattern.
// inspired from Mr. Matt Gallagher's idea, thanks!
//
#define xx 65
static const unsigned char base64DecodeLookup[256] =
{
	xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, 
	xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, 
	xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, 62, xx, xx, xx, 63, 
	52, 53, 54, 55, 56, 57, 58, 59, 60, 61, xx, xx, xx, xx, xx, xx, 
	xx,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 
	15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, xx, xx, xx, xx, xx, 
	xx, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 
	41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, xx, xx, xx, xx, xx, 
	xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, 
	xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, 
	xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, 
	xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, 
	xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, 
	xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, 
	xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, 
	xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, xx, 
};

@implementation Base64

/**
 * encode.
 * @param[in] data the data to encode.
 * @return encoded data.
 */
+ (NSData *)encode:(NSData *)data
{
	// prepare buffur to encode.
	NSInteger toEncodeLen = [data length];
	const char *toEncode = (char *) [data bytes];

	// prepare buffer to be encoded.
	NSInteger encodedLen = (toEncodeLen * 4 / 3 + 3) & ~0x3;
	NSMutableData* encoded = [[[NSMutableData alloc] initWithLength:encodedLen] autorelease];
	char *encodedData = (char *) [encoded bytes];
	NSUInteger putCnt = 0;	// actually put data.

	// devide into main loop and remainder loop.
	NSUInteger mainLoop = toEncodeLen / 3;
	const NSUInteger remnLoop = toEncodeLen % 3;
	NSUInteger main3Loop = 0;	// 3 byte access loop.
	if (mainLoop % sizeof(u_int32_t)) {
		// 4 byte acccess is faster than 3 byte access.
		// so we  will do 4 byte access as much as we can.
		main3Loop++;
		mainLoop--;
	}

	// mail loop.
	u_int32_t cell;
	NSUInteger idx;
	for (NSUInteger i = 0; i < mainLoop; i++) {
		cell = ntohl(*((u_int32_t *) toEncode));	// 4byte access.
		toEncode += 3;
		// 3 bytes to 4 bytes.
		for (NSUInteger j = 0; j < 4; j++) {
			// devide each 6 bit.
			idx = (cell >> bit_sfr[j]) & 0x3f;
			*encodedData = base64Str[idx];
			*encodedData++;
		}
		putCnt += 4;
	}
	// 3 byte accesss.
	if (main3Loop > 0) {
		// he alos makes 4 bytes.
		cell = 0;
		memcpy(&cell, toEncode, 3);	// 3 byte access.
		cell = ntohl(cell);
		toEncode += 3;
		// 3 bytes to 4 bytes.
		for (NSUInteger j = 0; j < 4; j++) {
			// devide each 6 bit.
			idx = (cell >> bit_sfr[j]) & 0x3f;
			*encodedData = base64Str[idx];
			*encodedData++;
		}
		putCnt += 4;
	}
	// reminder loop.
	if (remnLoop > 0) {
		cell = 0;
		memcpy(&cell, toEncode, remnLoop);	// less than 3 byte access.
		cell = ntohl(cell);
		toEncode += remnLoop;
		for (NSUInteger j = 0; j < (remnLoop + 1); j++) {
			// devide each 6 bit.
			idx = (cell >> bit_sfr[j]) & 0x3f;
			*encodedData = base64Str[idx];
			*encodedData++;
			putCnt++;
		}
		// must padding with '=' till no reminder with divde 4.
		while ((putCnt % 4) != 0) {
			*encodedData = padChar;
			*encodedData++;
			putCnt++;
		}
	}

	[encoded setLength:putCnt];
	return encoded;
}

/**
 * decode.
 * @param[in] data deta to decode.
 * @return decoded data.
 */
+ (NSData *)decode:(NSData *)data
{
	// encoded data.
	const u_int32_t *const encTop = (u_int32_t*) [data bytes];
	const u_int32_t *encData = encTop;
	const NSInteger encLen = [data length];
	if (encLen % 4 != 0) {
		// it's error, of cause.
		return nil;
	}

	// prepare for decoded data.
	const NSUInteger decLenTemp = encLen / 4 * 3 + sizeof(u_int32_t);	// little bit larger than encoded / 4 * 3.
	NSMutableData *const decoded = [[[NSMutableData alloc] initWithLength:decLenTemp] autorelease];
	// prepare pointer.
	char *const topData = (char *) [decoded bytes];
	char *decodedData = topData;
	NSUInteger putCnt = 0;	// actually put data.

	// devide into main loop and remainder loop.
	const NSInteger mainLoop = encLen / sizeof(u_int32_t);

	u_int32_t b64ch = 0;
	// main loop.
	u_int32_t cell, write_cell;
	for (NSInteger i = 0; i < (mainLoop - 1); i++) {
		cell = htonl(*((u_int32_t*) encData));	// 4byte access.
		encData++;
		// 4 bytes to 3 bytes.
		write_cell = 0;
		for (NSUInteger j = 0; j < 4; j++) {
			write_cell <<= 6;
			// lookup base64 character.
			b64ch = base64DecodeLookup[(cell & 0xff000000) >> 24];
			write_cell |= b64ch;
			cell <<= 8;
		}
		putCnt += 3;

		write_cell <<= 8;
		write_cell = htonl(write_cell);
		*((u_int32_t *) decodedData) = write_cell;	// 4 byte access, but 3 byte is required.
		decodedData += 3;
	}

	// i do not know why, but some problem with Xcode...
	unsigned char last4[sizeof(u_int32_t)];
	for (NSInteger i = (mainLoop - 1); i < mainLoop; i++) {
		memcpy(last4, encData, sizeof(u_int32_t));	// 4byte access.
		// only for intel architecture.
		encData++;
		// 4 bytes to 3 bytes.
		write_cell = 0;
		NSUInteger j = 0;
		for (; j < 4; j++) {
			// lookup base64 character.
			b64ch = (u_int32_t) base64DecodeLookup[last4[j]];
			if (b64ch == xx) {
				//NSLog(@"break at end");
				break;
			}
			write_cell <<= 6;
			write_cell = write_cell | b64ch;
		}
		NSUInteger byteWrite = 3;
		switch (j) {
			case 3:	// 1 '=' found.
				byteWrite = 2;
				write_cell <<= (24 - 18 + 8);	// because expresses 24 bit with 18 bit. and (32 - 8) value too.
				break;
			case 2:	// 2 '=' found.
				byteWrite = 1;
				write_cell <<= (16 - 12 + 16);	// becase expresses 16 bit with 12 bit. and (32 - 16) value too.
				break;
			default:
				// no '=' found.
				// 3 '=' never comes.
				write_cell <<= 8;	// expresses 24 bit full. but give her a (32 - 24) value.
				break;
		}
		putCnt += byteWrite;

		write_cell = htonl(write_cell);
		*((u_int32_t *) decodedData) = write_cell;	// non required area will cut off later.
	}

	[decoded setLength:putCnt];
	return decoded;
}

@end
