/**@file
 * ChangeDirectoryDialog.m
 * @brief Change directory dialog box class.
 * @author toru kageyama
 * Created by Toru Kageyama on 10/10/22.
 * Copyright 2010 Comona Co., Ltd. All rights reserved.
 */

#import "ChangeDirectoryDialog.h"
#import "MyDocument.h"
#import "ViewController.h"
#import "FileDisplayController.h"
#import "MyUtil.h"

@implementation ChangeDirectoryDialog

// properties.
@synthesize m_currentDirectry;

/**
 * Base class retrieves sheet pointer.
 * @return sheet ID.
 */
- (id)sheet
{
	// initialize controls.
	[m_dirName setStringValue:m_currentDirectry];
	[MyUtil focusTo:m_dirName onDialog:self];
	return [[m_document controller] cdDialog];
}

/**
 * OK button has pressed.
 */
- (void)onOk
{
	id fd = [self fileDisplay];
	if (fd) {
		[[m_document controller] directoryDoubleClicked:fd];
	}
}

/**
 * Ask can ok.
 */
- (BOOL)canOk
{
	id fd = [self fileDisplay];
	if (fd) {
		return YES;
	}
	else {
		[MyUtil messageBoxInternational:@"dir not found"];
		[MyUtil focusTo:m_dirName onDialog:self];
		return NO;
	}

}

/**
 * get file display.
 * @return file display object, nil if error.
 */
- (id)fileDisplay
{
	id controller = [m_document controller];
	id fdc = [controller fileDisplayController];
	NSString *path = [m_dirName stringValue];
	if (([path length] > 0) && ([path compare:@"/"] != NSOrderedSame)) {
		if ([path characterAtIndex:[path length] - 1] == '/') {
			path = [path substringToIndex:[path length] - 1];
		}
	}

	id fd = [fdc retrieveFileDisplay:path];
	if ([fd isDirectory] == NO) {
		fd = nil;	// physical file.
	}
	return fd;
}

// constructor.
- (id)init
{
	self = [super init];
	if (self) {
		m_currentDirectry = nil;
	}
	return self;
}

// destructor.
- (void)dealloc
{
	[m_currentDirectry release];
	[super dealloc];
}

@end
