/**@file
 * ControlDisabler.m
 * @brief Disable controls.
 * @author toru kageyama
 * Created by Toru Kageyama on 10/09/22.
 * Copyright 2010 Comona Co., Ltd. All rights reserved.
 */

#import "ControlDisabler.h"
#import "MyDocument.h"
#import "ViewController.h"
#import "ShsslAppDelegate.h"
#import "MenuEnabler.h"

@implementation ControlDisabler

// properties.
@synthesize m_canClose;

/**
 * set disable button.
 * @param[in] button buton to disalbe.
 */
- (void)disalbeButton:(NSButton *)button
{
	[m_buttons addObject:button];
	[button setEnabled:NO];
}

/**
 * set textfields disable to edit.
 * @param[in] textField text field to disalbe.
 */
- (void)disableTextField:(NSTextField *)textField
{
	[m_textFields addObject:textField];
	[textField setEditable:NO];
}

/**
 * disable frame close style.
 * @param[in] frame window to close style mask.
 */
- (void)disableWindowCloseStyle:(NSWindow *)frame
{
	[m_frames addObject:frame];
	NSUInteger mask = [frame styleMask];
	mask ^= NSClosableWindowMask;
	[frame setStyleMask:mask];
}

/**
 * reset all.
 */
- (void)resetAll
{
	for (id txt in m_textFields) {
		[txt setEditable:YES];
	}
	[m_textFields removeAllObjects];
	for (id but in m_buttons) {
		[but setEnabled:YES];
	}
	[m_buttons removeAllObjects];
	for (id frm in m_frames) {
		NSUInteger mask = [frm styleMask];
		mask |= NSClosableWindowMask;
		[frm setStyleMask:mask];
	}
	[m_frames removeAllObjects];
	[self setCanClose:YES];
}

/**
 * set can can close or not.
 * @param[in] close YES, if document can close, NO id cannot.
 */
- (void)setCanClose:(BOOL)close
{
	m_canClose = close;

	// enable/disable quit menu.
	BOOL cannotCloseAll = NO;
	NSDocumentController *const docCon = [NSDocumentController sharedDocumentController];
	NSArray *const documents = [docCon documents];
	for (id doc in documents) {
		id controller = [doc controller];
		ControlDisabler *const ctrl = [controller controlDisabler];
		if (ctrl->m_canClose == NO) {
			cannotCloseAll = YES;
			break;
		}
	}
	// to tell an app delegate, application can quit or no.
	id deleg = [NSApp delegate];
	id menuEnabler = [deleg menuEnabler];
	[menuEnabler setCanExit:cannotCloseAll == NO ? YES : NO];
}

// constructor.
- (id)init
{
	self = [super init];
	if (self) {
		m_buttons = [[NSMutableArray alloc] init];
		m_textFields = [[NSMutableArray alloc] init];
		m_canClose = YES;
	}
	return self;
}

// destructor.
- (void)dealloc
{
	[m_buttons release];
	[m_textFields release];
	[super dealloc];
}

@end
