/**@file
 * FileDisplayCollection.h
 * @brief Collection class for FileDisplay.
 * @author toru kageyama
 * Created by Toru Kageyama on 10/10/08.
 * Copyright 2010 Comona Co., Ltd. All rights reserved.
 */

#import <Cocoa/Cocoa.h>
#import "FileDisplay.h"

/**
 * Collection class for FileDisplay.
 */
@interface FileDisplayCollection : FileDisplay <NSTableViewDataSource, NSTableViewDelegate> {
@private
	NSMutableArray *m_displays;				// FileDisplay array.
	NSMutableDictionary *m_invalidates;		// FileDisplay invalidatings.
	id m_controller;						// ViewController.
	NSString *m_fullPath;					// full path name.
	NSMutableDictionary *m_pathDictionary;	// path dictionary.
	// editing support.
	NSString *m_identifierTagOnEdit;		// column identifier on edit.
	NSString *m_columnValueOnEdit;			// column old value on edit.
	// sort values.
	NSString *m_sortKey;					// sort key identifier.
	BOOL m_ascending;						// ascending key.
	// sorted array for display.
	NSMutableArray *m_sotedArray;			// sorted with user assigned type clicked table column header.
//	BOOL m_dummy;							// dummy flag.
}
// properties.
@property (nonatomic, assign, setter = setController, getter = controller) id m_controller;
@property (nonatomic, assign, getter = fullPath, readonly) NSString *m_fullPath;
@property (nonatomic, retain, setter = setIdentifierTagOnEdit, getter = identifierTagOnEdit) NSString *m_identifierTagOnEdit;
@property (nonatomic, retain, setter = setColumnValueOnEdit, getter = columnValueOnEdit) NSString *m_columnValueOnEdit;
@property (nonatomic, retain, setter = setSortKey, getter = sortKey) NSString *m_sortKey;
@property (nonatomic, assign, setter = setAscending, getter = isAscending) BOOL m_ascending;
//@property (nonatomic, assign, setter = setDummy, getter = isDummy) BOOL m_dummy;

// set full path name.
- (void)setFullPath:(NSString *)fullPath;
// get object count.
- (NSUInteger)count;
// get directory count.
- (NSUInteger)directoryCount;
// get object.
- (id)objectAtIndex:(NSUInteger)index;
// get display object count.
- (NSUInteger)displayCount;
// display object at index.
- (id)displayObjectAtIndex:(NSUInteger)index;
// get directory.
- (id)directoryAtIndex:(NSUInteger)index;
// remove all objects.
- (void)removeAllObjects;
// remove one item.
- (void)removeObjectAtIndex:(NSUInteger)index;
// remove one item of name.
- (void)removeObjectOfName:(NSString *)fileName;
// add display line.
- (void)addFileDisplay:(NSString *)ftpResult;
// before add files.
- (void)invalidate;
// after add files.
- (void)updateAll;
// create from file display.
+ (FileDisplayCollection *)fileDisplayCollectionWithFileDisplay:(FileDisplay*)fileDisplay;
// set path dictionary.
- (void)setPathDictionary:(NSMutableDictionary *)pathDictionary;
// expands only current directory.
- (void)expandOutlineView:(NSOutlineView *)outlineView currentDirectory:(NSString *)currentDir;
// copy children.
- (void)copyChildren:(id)src;
// ask path is exist.
- (BOOL)hasPath:(NSString *)path isDir:(BOOL*)isDir;
// ask path is exist, internal version.
- (BOOL)internalHasPath:(NSString *)path isDir:(BOOL*)isDir intrenalPath:(NSString *)intrenalPath;
// get file display.
- (id)fileDisplay:(NSString *)pathName;
// get file display.
- (id)fileDisplay:(NSArray *)pathNames depth:(NSUInteger)depth;
// accepts drop or not.
- (BOOL)tableView:(NSTableView *)tableView acceptDrop:(id <NSDraggingInfo>)info row:(int)row dropOperation:(NSTableViewDropOperation)operation;
// accepts drop or not into the row.
- (NSDragOperation)tableView:(NSTableView *)tableView validateDrop:(id <NSDraggingInfo>)info proposedRow:(int)row proposedDropOperation:(NSTableViewDropOperation)operation;
// prepare drag operation and copy the file name to the paste board.
- (BOOL)tableView:(NSTableView *)tableView writeRows:(NSArray *)rows toPasteboard:(NSPasteboard *)pboard;
// children sort.
- (void)childrenSort;
// add dummy directory for a while.
- (void)addDummyDirectory:(NSString *)dirName;
@end
