/**@file
 * FileNameFormatter.m
 * @brief The formatter for remove link target at editing.
 * @author toru kageyama
 * Created by Toru Kageyama on 10/10/23.
 * Copyright 2010 Comona Co., Ltd. All rights reserved.
 */

#import "FileNameFormatter.h"
#import "MyUtil.h"

// invalid chars.
static NSString *const g_invalidStrings[] = { @"/", @"*", @"?", @";", @"->" };

@implementation FileNameFormatter

/**
 * normal string format.
 * @return file name with link target if exit.
 */
- (NSString *)stringForObjectValue:(id)obj
{
	return [NSString stringWithFormat:@"%@", obj];
}

/**
 * editing time string format.
 * @return file name without link target if exit.
 */
- (NSString *)editingStringForObjectValue:(id)obj
{
	NSString *fmt = [NSString stringWithFormat:@"%@", obj];
	const NSRange range = [fmt rangeOfString:@" -> "];
	if (range.length > 0) {
		fmt = [fmt substringToIndex:range.location];
	}
	return fmt;
}
/**
 * input verifier.
 */
- (BOOL) isPartialStringValid:(NSString **)partialStringPtr
		proposedSelectedRange:(NSRangePointer)proposedSelRangePtr
			   originalString:(NSString *)origString
		originalSelectedRange:(NSRange)origSelRange
			 errorDescription:(NSString **)error
{
	if ([FileNameFormatter isValidFileName:*partialStringPtr] == NO) {
		NSBeep();
		*error = @"Invalid string input";
		return NO;
	}
	*error = nil;
	return YES;
}

/**
 * returs error occurs or not.
 * @param[out] anObject object value reciever.
 * @param[in] user input.
 * @param[out] error receiver.
 * @return YES if no error, NO if error.
 */
- (BOOL)getObjectValue:(id *)anObject forString:(NSString *)string errorDescription:(NSString **)error
{
	*anObject = [NSString stringWithFormat:@"%@", string];

	if ([string rangeOfString:@"/"].length == 0) {
		return YES;
	}
	else {
		return NO;
	}
}

/**
 * is valid file name?
 * @param[in] fileName file name of user input.
 * @return YES if valid input, NO if invalid input.
 */
+ (BOOL)isValidFileName:(NSString *)fileName
{
	for (NSUInteger i = 0; i < COUNT_OF(g_invalidStrings, NSString *); i++) {
		if ([fileName rangeOfString:g_invalidStrings[i]].length > 0) {
			return NO;
		}
	}
	return YES;
}

@end
