/**@file
 * NumberOnlyFormatter.m
 * Thanks for <http://borkware.com/quickies/single?id=93>
 * @brief class o limit user input 0 - 9.
 * @author toru kageyama
 * Created by Toru Kageyama on 10/09/02.
 * Copyright 2010 Comona Co., Ltd. All rights reserved.
 */

#import "NumberOnlyFormatter.h"

@implementation NumberOnlyFormatter

/**
 * input verifier.
 */
- (BOOL) isPartialStringValid:(NSString **)partialStringPtr
		proposedSelectedRange:(NSRangePointer)proposedSelRangePtr
			   originalString:(NSString *)origString
		originalSelectedRange:(NSRange)origSelRange
			 errorDescription:(NSString **)error
{
	NSCharacterSet *const nonDigits = [[NSCharacterSet decimalDigitCharacterSet] invertedSet];
	const NSRange newStuff = NSMakeRange(origSelRange.location, proposedSelRangePtr->location - origSelRange.location);
	NSString *const newStuffString = [*partialStringPtr substringWithRange:newStuff];

	if ([newStuffString rangeOfCharacterFromSet:nonDigits
										options:NSLiteralSearch].location != NSNotFound) {
		NSBeep();
		*error = @"Input is not an integer";
		return NO;
	}
	else {
		*error = nil;
		return YES;
	}
}

/**
 * normal string format.
 * @return file name with link target if exit.
 */
- (NSString *)stringForObjectValue:(id)obj
{
	return [NSString stringWithFormat:@"%@", obj];
}

@end
