/**@file
 * SendRecvThread.m
 * @brief Taking time communication will done by this thread calss.
 * @author toru kageyama
 * Created by Toru Kageyama on 10/10/07.
 * Copyright 2010 Comona Co., Ltd. All rights reserved.
 */

#import "SendRecvThread.h"
#import "ViewController.h"

@implementation SendRecvThread

// properties.
@synthesize m_param0;
@synthesize m_param1;

/**
 * start thread.
 * @param[in] procId process id.
 * @param[in] controller view controller.
 */
- (void)startThread:(NSUInteger)procId controller:(id)controller
{
	m_proc = procId;
	m_controller = controller;

	// create thread.
	[NSThread detachNewThreadSelector:@selector(threadProc:) toTarget:self withObject:nil];
}

/**
 * thread procedure.
 * @param[in] anArgument arugment from parent thread.
 */
- (void)threadProc:(id)anArgument
{
	NSAutoreleasePool *const pool = [[NSAutoreleasePool alloc] init];

	@try {
		// select procesure.
		switch (m_proc) {
			case PROCESS_ID_RETRIEVE_FILE:
				[m_controller retrieveFileSubThread];
				break;
			case PROCESS_ID_RETRIEVE_DROPPED_FILE:
				[m_controller retrieveDroppedFileSubThread:m_param0];
				break;
			case PROCESS_ID_ADD_FILE:
				[m_controller addFileSubThread];
				break;
			case PROCESS_ID_PASTE_FILE:
				[m_controller pasteSubThread];
				break;
			case PROCESS_ID_PASTE_FILE_FINDER:
				[m_controller pasteFromFinderSubThread];
				break;
			case PROCESS_ID_DROPPED_FILE_FINDER:
				[m_controller droppedFromFinderSubThread:m_param0 fileDisplay:m_param1];
				break;
			default:
				break;
		}
	}
	@catch (NSException *e) {
	}
	@finally {
		// close dialog box.
		[m_controller endSubProcess];
		[pool drain];
	}
}

// constructor.
- (id)init
{
	self = [super init];
	if (self) {
		m_proc = PROCESS_ID_UNKNOWN;
		m_controller = nil;
		m_param0 = nil;
		m_param1 = nil;
	}
	return self;
}

// destructor
- (void)dealloc
{
	[m_param0 release];
	[m_param1 release];
	[super dealloc];
}

@end
