/**@file
 * ViewController.m
 * @brief The view controller class.
 * @author toru kageyama
 * Created by Toru Kageyama on 10/10/07.
 * Copyright 2010 Comona Co., Ltd. All rights reserved.
 */

#import "ViewController.h"
#import "MyDocument.h"
#import "ConnectSettings.h"
#import "MyUtil.h"
#import "ShsslClient.h"
#import "FileDisplayCollection.h"
#import "FileDisplayController.h"
#import "MyBaseDialog.h"
#import "RenameDialog.h"
#import "AskSkipOverwriteDialog.h"
#import "FileDropTableView.h"
#import "ShsslAppDelegate.h"
#import "CopySource.h"
#import "SendRecvThread.h"
#import "ProcessingDialog.h"
#import "ControlDisabler.h"

@implementation ViewController

// properties.
@synthesize m_owner;
@synthesize m_mainWindow;
@synthesize m_settingDialog;
@synthesize m_renameDialog;
@synthesize m_chmodDialog;
@synthesize m_fileDisplayController;
@synthesize m_askSkipOverwriteDialog;
@synthesize m_cdDialog;
@synthesize m_client;
@synthesize m_savedCurDir;
@synthesize m_cancel;
@synthesize m_progressTitle;
@synthesize m_controlDisabler;
@synthesize m_fileView;

/**
 * connect button.
 * @param[in] sender sender object. connect button.
 */
- (IBAction)connectButton:(id)sender
{
	[m_statusField setStringValue:@""];

	if ([m_client isConnect] == NO) {
		// connect.
		id setting = [m_owner connectionSetting];
		if ([[setting urlString] length] == 0) {
			[MyUtil messageBoxInternational:@"No connect settings"];
			return;
		}
		if ([m_client connect:YES] != NO) {
			[m_connectButton setTitle:NSLocalizedString(@"Disconnect", @"Disconnect")];
			[m_settingButton setEnabled:NO];
			[m_fileView setDataSource:[m_fileDisplayController currentFileViewDataSource]];
			[m_fileView setDelegate:[m_fileDisplayController currentFileViewDataSource]];
			// update status bar.
			id item = [m_fileDisplayController currentFileViewDataSource];
			NSString *const status = [NSString stringWithFormat:@"%@ [%d files]", [item fullPath], [item displayCount]];
			[m_statusField setStringValue:status];
		}
		else {
			[MyUtil messageBoxInternational:@"connect failed"];
		}
		[m_fileView reloadData];
		[self resetDirectoryView];
	}
	else {
		[self disconnected];
	}
}

/**
 * cancel button.
 * @param[in] sender sender object. connect button.
 */
- (IBAction)cancelButton:(id)sender
{
	m_cancel = YES;
}

/**
 * rename menu.
 * @param[in] sender the sender object.
 */
- (void)renameToMenu:(id)sender
{
	[m_renameDialogController setProcess:RENAME_DIALOG_RENAME];
	[m_renameDialogController openDialog:sender];
}

/**
 * chmod menu.
 * @param[in] sender the sender object.
 */
- (void)chmodToMenu:(id)sender
{
	[m_chmodDialogController openDialog:sender];
}

/**
 * rmv menu.
 * @param[in] sender the sender object.
 */
- (void)removeFileMenu:(id)sender
{
	NSString *const askRemove = NSLocalizedString(@"are you remove", @"are you remove");
	if ([MyUtil askYesNo:askRemove] == 1) {
		// get data source.
		id ds = [m_fileView dataSource];

		// try to remove.
		NSIndexSet *const indexSet = [m_fileView selectedRowIndexes];
		NSUInteger index = [indexSet firstIndex];
		NSUInteger firstIndex = index;
		NSUInteger removed = 0;
		while (index != NSNotFound) {
			// file display.
			id fd = [ds displayObjectAtIndex:index - removed];	// get FileDisplay from FileDisplayCollection.
			// remove file.
			[m_fileDisplayController setCurrentDirectory:[ds fullPath]];
			[m_client setCurrentDirectory:[ds fullPath]];
			if ([self rmvFile:[fd fileTitle] isDirectory:[fd isDirectory]] == NO) {
				NSString *const msg = [NSString stringWithFormat:NSLocalizedString(@"remove failed", :@"remove failed"), [fd fileName]];
				[MyUtil messageBox:msg];
				break;
			}
			// remove file from data source FileDisplayCollection.
			NSString *const fileName = [fd fileName];
			[fileName retain];
			[ds removeObjectOfName:fileName];
			[fileName release];

			index = [indexSet indexGreaterThanIndex:index];
			removed++;
		}

		[self directorySelectionHasChanged:ds];
		// set first index to new new selected file.
		const NSUInteger newCnt = [ds count];
		if (firstIndex >= newCnt) {
			if (newCnt > 0) {
				firstIndex = newCnt - 1;
			}
			else {
				firstIndex = -1;
			}
		}
		[self setFileCursorPosition:firstIndex];
		[m_fileView reloadData];
		[m_directoryView reloadData];
	}
}

/**
 * mkdir menu.
 * @param[in] sender the sender object.
 */
- (void)makeDirectoryMenu:(id)sender
{
	m_cancel = NO;	// just in case...
	[m_renameDialogController setProcess:RENAME_DIALOG_MKDIR];
	[m_renameDialogController openDialog:sender];
}

/**
 * retr menu.
 * @param[in] sender the sender object.
 */
- (void)retrieveFile:(id)sender
{
	[self longProcessStart];

	// sub thread class.
	[m_progressThread release];
	m_progressThread = [[SendRecvThread alloc] init];
	[m_progressThread startThread:PROCESS_ID_RETRIEVE_FILE controller:self];
}

/**
 * retrieve file driven by sub thread.
 */
- (void)retrieveFileSubThread
{
	// selected row indice.
	NSIndexSet *const indexSet = [m_fileView selectedRowIndexes];
	NSUInteger index = [indexSet firstIndex];
	id ds = [m_fileView dataSource];	// getting data source.
	while (index != NSNotFound) {
		// file display.
		id fd = [ds displayObjectAtIndex:index];	// get FileDisplay from FileDisplayCollection.
		// retrieve file.
		if ([fd isDirectory] == NO) {
			// physical file.
			@try {
				[m_client setCurrentDirectory:[m_fileDisplayController currentDirectory]];
				[self retrieveFileContents:fd pathTo:nil];
			}
			@catch (NSException *e) {
				NSString *const msg = [NSString stringWithFormat:NSLocalizedString(@"retrieve failed", :@"retrieve failed"), [fd fileName]];
				[MyUtil messageBox:msg];
				break;
			}
		}
		else {
			// directory.
			@try {
				[m_client setCurrentDirectory:[fd fullPath]];
				if ([self saveDirectory:fd path:nil] == NO) {
					@throw [NSException exceptionWithName:@"retr" reason:@"save directory failed" userInfo:nil];
				}
			}
			@catch (NSException *e) {
				NSString *const msg = [NSString stringWithFormat:NSLocalizedString(@"retrieve failed", :@"retrieve failed"), [fd fileName]];
				[MyUtil messageBox:msg];
				break;
			}
			@finally {
				[m_client setCurrentDirectory:[ds fullPath]];
			}
		}
		
		// get next index.
		index = [indexSet indexGreaterThanIndex:index];
		if ([self isCanceled]) {
			break;	// user abort.
		}
	}
}

/**
 * retr by drag and drop.
 * @param[in] targetDir drop target directory.
 */
- (void)retrieveDroppedFile:(NSString *)targetDir
{
	[self longProcessStart];

	// sub thread class.
	[m_progressThread release];
	m_progressThread = [[SendRecvThread alloc] init];
	[m_progressThread setParam0:targetDir];
	[m_progressThread startThread:PROCESS_ID_RETRIEVE_DROPPED_FILE controller:self];
}

/**
 * retr by drag and drop driven by sub thread.
 * @param[in] targetDir drop target directory.
 */
- (void)retrieveDroppedFileSubThread:(NSString *)targetDir
{
	// Get the finder path out of the URL here;
	id ds = [m_fileView dataSource];
	NSIndexSet *const indexSet = [m_fileView selectedRowIndexes];
	NSInteger index = [indexSet firstIndex];
	NSString *localPath = nil;

	while (index != NSNotFound) {
		id fd = [ds displayObjectAtIndex:index];
		localPath = [targetDir stringByAppendingPathComponent:[fd fileTitle]];
		if ([fd isDirectory]) {
			// directory.
			[m_client setCurrentDirectory:[fd fullPath]];
			if ([self saveDirectory:fd path:localPath] == NO) {
				break;
			}
		}
		else {
			// physical directory.
			@try {
				if ([self isSkipFile:localPath forDirectory:NO] == NO) {
					[m_owner setLastDirectory:targetDir];
					[m_client setCurrentDirectory:[ds fullPath]];
					[self retrieveFileContents:fd pathTo:localPath];
				}
			}
			@catch (NSException *e) {
				NSString *const msg = [NSString stringWithFormat:NSLocalizedString(@"retrieve failed", :@"retrieve failed"), [fd fileName]];
				[MyUtil messageBox:msg];
				break;
			}
		}
	
		// get next index.
		index = [indexSet indexGreaterThanIndex:index];
		if ([self isCanceled]) {
			break;	// user abort.
		}
	}

	if ([m_client isConnect]) {
		[m_client setCurrentDirectory:[ds fullPath]];
	}
}

/**
 * retrieve file.
 * @param[in] fd file display object which is the target of retrive.
 * @param[in] pathTo local path name to store. if nil ask with dialog box.
 */
- (void)retrieveFileContents:(id)fd pathTo:(NSString *)pathTo
{
	// physical file.
	NSMutableData *fileContents = nil;
	@try {
		// update dialog box title.
		[self updateProgressTitle:[fd fileTitle] directory:[m_client currentDirectory] withFormat:@"Fetching:"];

		fileContents = [[NSMutableData alloc] init];
		if ([m_client retr:[fd fileTitle] content:fileContents] == NO) {
			@throw [NSException exceptionWithName:@"retr" reason:@"RETR failed" userInfo:nil];
		}
		if ([self isCanceled]) {
			return;	// user abort.
		}

		// update dialog box title.
		[self updateProgressTitle:[fd fileTitle] directory:pathTo withFormat:@"Saving:"];

		// save into the file.
		if ([self saveFileContent:fileContents as:fd lastDirectory:YES pathTo:pathTo] == NO) {
			@throw [NSException exceptionWithName:@"retr" reason:@"filesave" userInfo:nil];
		}
	}
	@catch (NSException *e) {
		@throw e;
	}
	@finally {
		[fileContents release];
	}
}

/**
 * stor menu.
 * @param[in] sender the sender object.
 */
- (void)addFile:(id)sender
{
	[self longProcessStart];

	// sub thread class.
	[m_progressThread release];
	m_progressThread = [[SendRecvThread alloc] init];
	[m_progressThread startThread:PROCESS_ID_ADD_FILE controller:self];
}

/**
 * stor menu driven by sub thread.
 */
- (void)addFileSubThread
{
	// choose file.
	NSOpenPanel *const openDlg = [NSOpenPanel openPanel];
	NSURL *const url = [NSURL URLWithString:[m_owner lastDirectory]];
	[openDlg setDirectoryURL:url];
	[openDlg setCanChooseDirectories:YES];
	if ([openDlg runModalForDirectory:[m_owner lastDirectory] file:@""] == NSOKButton) {
		// update display.
		id ds = [m_fileView dataSource];
		[self silentDirectorySelectionHasChanged:ds];
		// get file path.
		NSArray *const filePaths = [openDlg filenames];
		for (NSUInteger i = 0; i < [filePaths count]; i++) {
			NSString *const filePath = [filePaths objectAtIndex:i];
			BOOL storResult = YES;
			if ([MyUtil isDirectory:filePath] == NO) {
				// store file.
				storResult = [self storeFile:filePath];
			}
			else {
				// store directory.
				storResult = [self storeDirectory:filePath];
			}
			if (storResult == NO) {
				[MyUtil messageBoxInternational:@"STOR failed"];
				break;
			}
			if ([self isCanceled]) {
				break;	// user abort.
			}
		}

		// update display.
		if ([m_client isConnect]) {
			[self directorySelectionHasChanged:ds];
			// select newly added file.
			[self selectAddedFiles:filePaths];
		}
	}
}

/**
 * stor menu with paste.
 * @param[in] sender the sender object.
 */
- (void)paste:(id)sender
{
	[self longProcessStart];

	// sub thread class.
	[m_progressThread release];
	m_progressThread = [[SendRecvThread alloc] init];
	[m_progressThread startThread:PROCESS_ID_PASTE_FILE controller:self];
}

/**
 * stor menu with paste driven by sub thread.
 * @param[in] sender the sender object.
 */
- (void)pasteSubThread
{
	id ds = [m_fileView dataSource];
	// get file name array.
	id delegate = [NSApp delegate];
	[self setSavedCurDir:[ds fullPath]];
	@try {
		NSArray *const fileNameArray = [delegate copyFileArray];
		for (CopySource *const cpSrc in fileNameArray) {
			if ([cpSrc copyTo:self] == NO) {	// copy.
				@throw [NSException exceptionWithName:@"paste" reason:@"unknown error" userInfo:nil];
			}

			if ([self isCanceled]) {
				break;	// user abort.
			}
		}
		// update display.
		[self directorySelectionHasChanged:ds];
		// select newly added file.
		[self selectAddedFiles:fileNameArray];
	}
	@catch (NSException *e) {
		[MyUtil messageBoxInternational:[e reason]];
	}
	@finally {
	}
}

/**
 * stor menu with paste from finder.
 * @param[in] sender the sender object.
 */
- (void)pasteFromFinder:(id)sender
{
	[self longProcessStart];

	// sub thread class.
	m_progressThread = [[SendRecvThread alloc] init];
	[m_progressThread startThread:PROCESS_ID_PASTE_FILE_FINDER controller:self];
}

/**
 * stor menu with paste from finder driven by sub thread.
 */
- (void)pasteFromFinderSubThread
{
	id ds = [m_fileView dataSource];

	// see the contnent of paste board.
	NSPasteboard *const pboard = [NSPasteboard generalPasteboard];
	//NSLog(@"NSFilenamesPboardType = %@", NSFilenamesPboardType);
	NSArray *const typeArray = [pboard propertyListForType:NSFilenamesPboardType];
	for (id filePath in typeArray) {
		//NSLog(@"found = %@", filePath);
		if ([MyUtil isDirectory:filePath] == NO) {
			// store file.
			if ([self storeFile:filePath] == NO) {
				break;
			}
		}
		else {
			// store directory.
			if ([self storeDirectory:filePath] == NO) {
				break;
			}
		}
		
		if ([self isCanceled]) {
			break;	// user abort.
		}
	}

	// update display.
	[self directorySelectionHasChanged:ds];
	// select newly added file.
	[self selectAddedFiles:typeArray];
}

/**
 * copy.
 * @param[in] sender the sender object.
 */
- (void)copy:(id)sender
{
	id appDelegate = [NSApp delegate];
	[appDelegate copyData:[self makeFileNameArray] isCut:NO ofController:self];
}

/**
 * cut.
 * @param[in] sender the sender object.
 */
- (void)cut:(id)sender
{
	id appDelegate = [NSApp delegate];
	[appDelegate copyData:[self makeFileNameArray] isCut:YES ofController:self];
}

/**
 * parent directory.
 * @param[in] sender the sender object.
 */
- (void)parentDirectory:(id)sender
{
	NSString *const curDir = [m_fileDisplayController currentDirectory];
	NSRange range = [curDir rangeOfString:@"/" options:NSBackwardsSearch];
	NSString *parentDir = nil;
	if (range.location < 1) {
		// root dir.
		parentDir = @"/";
	}
	else {
		parentDir = [curDir substringToIndex:range.location];
	}

	id fd = [m_fileDisplayController retrieveFileDisplay:parentDir];
	if (fd) {
		[self directoryDoubleClicked:fd];
	}
}

/**
 * change directory.
 * @param[in] sender the sender object.
 */
- (IBAction)changeDirectory:(id)sender
{
	[m_cdDialogController setCurrentDirectory:[m_fileDisplayController currentDirectory]];
	[m_cdDialogController openDialog:sender];
}

/**
 * make file name array.
 * @return file name array.
 */
- (NSArray *)makeFileNameArray
{
	NSMutableArray *const array = [NSMutableArray array];
	NSIndexSet *const indexSet = [m_fileView selectedRowIndexes];
	NSString *localPath = nil;
	id ds = [m_fileView dataSource];
	NSString *const targetDir = [ds fullPath];

	NSUInteger index = [indexSet firstIndex];
	while (index != NSNotFound) {
		id fd = [ds displayObjectAtIndex:index];
		localPath = [targetDir stringByAppendingPathComponent:[fd fileTitle]];
		[array addObject:localPath];

		// get next index.
		index = [indexSet indexGreaterThanIndex:index];
	}
	return array;
}

/**
 * reload menu.
 * @param[in] sender the sender object.
 */
- (void)reloadMenu:(id)sender
{
	m_cancel = NO;	// just in case...
	id ds = [m_fileView dataSource];
	[self directorySelectionHasChanged:ds];
}

/**
 * save directory.
 * @param[in] directoryDisplay directory display object.
 * @param[in] promptPath display dialog if nil.
 * @return YES if sccuess, NO if failed.
 */
- (BOOL)saveDirectory:(id)directoryDisplay path:(NSString *)promptPath
{
	// LIST again.
	const BOOL ret = [self fullpathListFileDisplay:directoryDisplay];
	if (ret == NO) {
		[MyUtil messageBoxInternational:@"LIST failed"];
		return NO;
	}

	// ask user for the save path.
	NSString *filePath = nil;
	if (promptPath == nil) {
		NSSavePanel *const saveDlg = [NSSavePanel savePanel];
		NSURL *const url = [NSURL URLWithString:[m_owner lastDirectory]];
		[saveDlg setDirectoryURL:url];
		if ([saveDlg runModalForDirectory:[m_owner lastDirectory] file:[directoryDisplay fileTitle]] == NSOKButton) {
			filePath = [saveDlg filename];

			// create directory.
			if ([ViewController createDirectory:filePath] == NO) {
				return NO;
			}

			// update last saved directory
			NSString *const fl = [filePath lastPathComponent];
			NSString *const lastDir = [filePath substringToIndex:[filePath length] - [fl length]];
			[m_owner setLastDirectory:lastDir];
		}
		else {
			// do not save the file.
			return YES;
		}
	}
	else {
		filePath = promptPath;
		if ([self isSkipFile:filePath forDirectory:YES] == YES) {
			return YES;
		}
		if ([ViewController createDirectory:filePath] == NO) {
			return NO;
		}
	}

	// enum children.
	for (NSUInteger i = 0; i < [directoryDisplay displayCount]; i++) {
		id fd = [directoryDisplay displayObjectAtIndex:i];
		if ([fd isDirectory] == NO) {
			// physical file.
			if ([self retrieveFile:fd withParent:filePath] == NO) {
				return NO;
			}
		}
		else {
			BOOL saveDir = NO;
			@try {
				// directory, so recursive LIST.
				[m_client setCurrentDirectory:[fd fullPath]];
				if ([self recursiveSaveDirectory:fd withParent:filePath] == NO) {
					return NO;
				}
				saveDir = YES;
			}
			@catch (NSException *e) {
				@throw e;
			}
			@finally {
				if (saveDir) {
					[m_client setCurrentDirectory:[directoryDisplay fullPath]];
				}
			}
		}

		if ([self isCanceled]) {
			break;	// user abort.
		}
	}

	return YES;
}

/**
 * retr menu.
 * @param[in] fd file to retrieve.
 * @param[in] parentDir local parent directory path name.
 * @return YES if sccuess, NO if failed.
 */
- (BOOL)retrieveFile:(id)fd withParent:(NSString *)parentDir
{
	m_cancel = NO;	// just in case...
	// new file name.
	NSString *const newFile = [parentDir stringByAppendingPathComponent:[fd fileTitle]];

	BOOL ret = NO;
	NSMutableData *fileContents = nil;
	@try {
		if ([self isSkipFile:newFile forDirectory:NO] != NO) {
			// users choice is skip.
			return YES;
		}

		// update dialog box title.
		[self updateProgressTitle:[fd fileTitle] directory:[m_client currentDirectory] withFormat:@"Fetching:"];

		fileContents = [[NSMutableData alloc] init];
		if ([m_client retr:[fd fileTitle] content:fileContents] == NO) {
			@throw [NSException exceptionWithName:@"retr" reason:@"RETR failed" userInfo:nil];
		}
		if ([self isCanceled]) {
			ret = YES;
			return YES;	// user abort.
		}

		// update dialog box title.
		[self updateProgressTitle:[fd fileTitle] directory:parentDir withFormat:@"Saving:"];

		// save content.
		if ([self saveFile:newFile content:fileContents] == NO) {
			@throw [NSException exceptionWithName:@"retr" reason:@"file save failed" userInfo:nil];
		}
		ret = YES;
	}
	@catch (NSException *e) {
		//NSLog(@"error reason = %@", [e reason]);
		[MyUtil messageBoxInternational:[e reason]];
	}
	@finally {
		[fileContents release];
	}
	return ret;
}

/**
 * dropped from finder.
 * @param[in] tableView an table view object operation drag and drop.
 * @param[in] info drag and drop information.
 * @return YES if accepts drop, NO if not.
 */
- (BOOL)droppedFromFinder:(id <NSDraggingInfo>)info fileDisplay:(id)fileDisplay
{
	NSPasteboard *const pboard = [info draggingPasteboard];
	if ([[pboard types] containsObject:NSFilenamesPboardType]) {
		[self longProcessStart];

		// sub thread class.
		[m_progressThread release];
		m_progressThread = [[SendRecvThread alloc] init];
		id pb = pboard;
		[m_progressThread setParam0:pb];
		[m_progressThread setParam1:fileDisplay];
		[m_progressThread startThread:PROCESS_ID_DROPPED_FILE_FINDER controller:self];

		return YES;
	}

	return NO;
}

/**
 * dropped from finder driven by sub thread.
 * @param[in] tableView an table view object operation drag and drop.
 * @param[in] info drag and drop information.
 */
- (void)droppedFromFinderSubThread:(NSPasteboard *)pboard fileDisplay:(id)fileDisplay
{
	NSFileManager *const fileManager = [NSFileManager defaultManager];

	NSMutableArray *const fileArray = [NSMutableArray array];
	id file = nil;
	NSEnumerator *const enumerator = [[pboard propertyListForType:NSFilenamesPboardType] objectEnumerator];
	while (file = [enumerator nextObject]) {
		//NSLog(@"file = %@", file);
		[fileArray addObject:file];	// add file name into array for selection change.
		BOOL isDir = NO;
		const BOOL exist = [fileManager fileExistsAtPath:file isDirectory:&isDir];
		if (exist) {
			if (isDir == NO) {
				// store file.
				if ([self storeFile:file] == NO) {
					[MyUtil messageBoxInternational:@"STOR failed"];
					break;
				}
			}
			else {
				// store directory.
				if ([self storeDirectory:file] == NO) {
					[MyUtil messageBoxInternational:@"STOR failed"];
					break;
				}
			}
		}
		if ([self isCanceled]) {
			break;	// user abort.
		}
	}
	// update view.
	if ([m_client isConnect]) {
		[self directorySelectionHasChanged:fileDisplay];
		[self updateDirectoryView];
		[self selectAddedFiles:fileArray];
	}
}

/**
 * recursive save directory.
 * @param[in] directoryDisplay directory display object.
 * @param[in] parentDir parent directory path.
 * @return YES if sccuess, NO if failed.
 */
- (BOOL)recursiveSaveDirectory:(id)directoryDisplay withParent:(NSString *)parentDir
{
	// LIST again.
	const BOOL ret = [self fullpathListFileDisplay:directoryDisplay];
	if (ret == NO) {
		[MyUtil messageBoxInternational:@"LIST failed"];
		return NO;
	}

	// building full path name.
	NSString *const newDir = [parentDir stringByAppendingPathComponent:[directoryDisplay fileTitle]];
	if ([self isSkipFile:newDir forDirectory:YES] == YES) {
		return YES;
	}
	if ([ViewController createDirectory:newDir] == NO) {
		return NO;
	}

	// enum children.
	for (NSUInteger i = 0; i < [directoryDisplay displayCount]; i++) {
		id fd = [directoryDisplay displayObjectAtIndex:i];
		if ([fd isDirectory] == NO) {
			// physical file.
			if ([self retrieveFile:fd withParent:newDir] == NO) {
				return NO;
			}
		}
		else {
			// directory, so recursive LIST.
			if ([self recursiveSaveDirectory:fd withParent:newDir] == NO) {	// recursive call.
				return NO;
			}
			// rewind current directory.
			[m_client setCurrentDirectory:[directoryDisplay fullPath]];
		}

		if ([self isCanceled]) {
			break;	// user abort.
		}
	}
	
	return YES;
}

/*
 * create directory.
 * @param[in] path directory path to create.
 * @return YES if success, NO if failed.
 */
+ (BOOL)createDirectory:(NSString *)path
{
	NSFileManager *const fileManager = [NSFileManager defaultManager];
	BOOL isDir = NO;
	BOOL exist = [fileManager fileExistsAtPath:path isDirectory:&isDir];
	if (exist) {
		if (isDir == NO) {
			// remove file now.
			if ([fileManager removeItemAtPath:path error:NULL]) {
				return NO;
			}
			exist = NO;
		}
	}
	if (exist == NO) {
		exist = [fileManager createDirectoryAtPath:path withIntermediateDirectories:NO attributes:nil error:NULL];
	}
	return exist;
}

/**
 * save file contents.
 * @param[in] content file content.
 * @param[in] fileDisplay FileDisplay to get original file name.
 * @param[in] lastDirSave if YES save last saved directory, if NO not.
 * @param[in] pathTo path name to store, if nil ask.
 * @return YES if sccuess, NO if failed.
 */
- (BOOL)saveFileContent:(NSData *)content as:(id)fileDisplay lastDirectory:(BOOL)lastDirSave pathTo:(NSString *)pathTo;
{
	NSString *filePath = nil;
	if (pathTo == nil) {
		NSSavePanel *const saveDlg = [NSSavePanel savePanel];
		NSURL *const url = [NSURL URLWithString:[m_owner lastDirectory]];
		[saveDlg setDirectoryURL:url];
		if ([saveDlg runModalForDirectory:[m_owner lastDirectory] file:[fileDisplay fileTitle]] != NSOKButton) {
			return YES;
		}
		filePath = [saveDlg filename];
	}
	else {
		filePath = pathTo;
	}

	// save content.
	if ([self saveFile:filePath content:content] == NO) {
		return NO;
	}

	// update last saved directory
	if (lastDirSave) {
		NSString *fl = [filePath lastPathComponent];
		NSString *const lastDir = [filePath substringToIndex:[filePath length] - [fl length]];
		[m_owner setLastDirectory:lastDir];
	}
	return YES;
}

/**
 * save file contents.
 * @param[in] path file path.
 * @param[in] content file content.
 * @return YES if sccuess, NO if failed.
 */
- (BOOL)saveFile:(NSString *)path content:(NSData *)content
{
	if ([[NSFileManager defaultManager] createFileAtPath:path contents:nil attributes:nil] == NO) {
		return NO;
	}

	NSFileHandle *const fileHandle = [NSFileHandle fileHandleForWritingAtPath:path];
	if (fileHandle == nil) {
		// maybe a no parmission right.
		return NO;
	}
	[fileHandle writeData:content];
	[fileHandle synchronizeFile];
	[fileHandle closeFile];

	return YES;
}

/**
 * set current directory.
 * @param[in] currentDirectory new current directory.
 */
- (void)setCurrentDirectory:(NSString *)currentDirectory
{
	[m_fileDisplayController setCurrentDirectory:currentDirectory];
}

/**
 * reset directory view.
 */
- (void)resetDirectoryView
{
	[m_directoryView reloadData];
	[m_fileDisplayController expandOutlineView:m_directoryView];
}

/**
 * directory selection has changed.
 * @param[in] item newly selected item.
 */
- (void)directorySelectionHasChanged:(id)item
{
	if ([m_client isConnect] == NO) {
		// not connected.
		return;
	}

	// change directory.
	const BOOL ret = [self fullpathListFileDisplay:item];
	if ((ret == NO) || ([m_client commandResult] == NO)) {
		[MyUtil messageBoxInternational:@"LIST failed"];
		[m_fileView setDataSource:nil];
	}
	else {
		[m_fileView setDataSource:item];
		[m_fileView scrollRowToVisible:0];
		// set cursor index 0.
		[self setFileCursorPosition:0];
	}
	[m_fileView reloadData];
	//[m_directoryView reloadData];

	// status bar update.
	if ([m_client isConnect]) {
		NSString *const status = [NSString stringWithFormat:@"%@ [%d files]", [item fullPath], [item displayCount]];
		[m_statusField setStringValue:status];
	}
	else {
		[m_statusField setStringValue:@""];
	}
}

/**
 * directory selection has changed silently.
 * @param[in] item newly selected item.
 */
- (void)silentDirectorySelectionHasChanged:(id)item
{
	if ([m_client isConnect] == NO) {
		// not connected.
		return;
	}

	// change directory.
	const BOOL ret = [self fullpathListFileDisplay:item];
	if ((ret == NO) || ([m_client commandResult] == NO)) {
		[MyUtil messageBoxInternational:@"LIST failed"];
		[m_fileView setDataSource:nil];
	}
}

/**
 * rename file.
 * @param[in] item target item.
 * @param[in] from old name.
 * @param[in] to new name.
 * @return YES if success, NO if not.
 */
- (BOOL)renameFile:(id)item from:(NSString *)from to:(NSString *)to
{
	if ([m_client isConnect] == NO) {
		// not connected.
		return NO;
	}
	//NSLog(@"renameFileFrom:%@ to:%@", from, to);
	// change directory.
	const BOOL ret = [m_client renameFileFrom:from to:to];
	if ((ret == NO) || ([m_client commandResult] == NO)) {
		[MyUtil messageBoxInternational:@"REN failed"];
	}
	else {
		// reorder.
		id ds = [m_fileView dataSource];
		const BOOL listResult = [self fullpathListFileDisplay:ds];
		if (listResult) {
			for (NSUInteger i = 0; i < [ds displayCount]; i++) {
				id fd = [ds displayObjectAtIndex:i];
				if ([to compare:[fd fileTitle]] == NSOrderedSame) {
					[self setFileCursorPosition:i];
					break;
				}
			}
		}
		else {
			[MyUtil messageBoxInternational:@"LIST failed"];
		}

	}
	// update view.
	[m_directoryView reloadData];

	return ret;
}

/**
 * rename current selected file.
 * @param[in] to new name.
 * @return YES if success, NO if not.
 */
- (BOOL)renameFileTo:(NSString *)to
{
	id ds = [m_fileView dataSource];
	const NSInteger sel = [m_fileView selectedRow];
	id fd = [ds displayObjectAtIndex:sel];
	if ([[fd fileName] compare:to] != NSOrderedSame) {
		// rename.
		const BOOL ret = [self renameFile:ds from:[fd fileTitle] to:to];
		if (ret != NO) {
			// update display.
			[m_fileView reloadData];
		}
		return ret;
	}
	else {
		return YES;
	}
}

/**
 * chmod file.
 * @param[in] item target item.
 * @param[in] pathName file path.
 * @param[in] mode mode in 3 digits.
 * @return YES if success, NO if not.
 */
- (BOOL)chmod:(id)item file:(NSString *)pathName mode:(NSString *)mode
{
	if ([m_client isConnect] == NO) {
		// not connected.
		return NO;
	}
	NSLog(@"chmod:%@ to:%@", pathName, mode);
	// change directory.
	const BOOL ret = [m_client chmod:pathName mode:mode];
	if ((ret == NO) || ([m_client commandResult] == NO)) {
		[MyUtil messageBoxInternational:@"CHMOD failed"];
	}
	[m_directoryView reloadData];
	
	return ret;
}

/**
 * chmod current selected file.
 * @param[in] to new mode.
 * @return YES if success, NO if not.
 */
- (BOOL)chmodFileTo:(NSString *)mode
{
	id ds = [m_fileView dataSource];
	const NSInteger sel = [m_fileView selectedRow];
	id fd = [ds displayObjectAtIndex:sel];
	if ([fd fileMode] != [mode intValue]) {
		// chmod.
		const BOOL ret = [self chmod:ds file:[fd fileName] mode:mode];
		if (ret != NO) {
			// update display.
			[fd setFileMode:[mode intValue]];
			[m_fileView reloadData];
		}
		return ret;
	}
	else {
		return YES;	// no change.
	}
}

/**
 * rmv file.
 * @param[in] path file path to delete.
 * @param[in] isDirectory YES if drectory, NO if not directory.
 * @return YES if success, NO if not.
 */
- (BOOL)rmvFile:(NSString *)path isDirectory:(BOOL)isDirectory
{
	BOOL ret = NO;
	if (isDirectory == NO) {
		// remove file.
		ret = [m_client rm:path];
	}
	else {
		// remove directory.
		ret = [m_client rmd:path];
	}
	return ret;
}

/**
 * mkd.
 * @param[in] dirName directory name to create.
 * @return YES if success, NO if not.
 */
- (BOOL)mkd:(NSString *)dirName
{
	const BOOL ret = [m_client mkd:dirName];
	if (ret == NO) {
		[MyUtil messageBoxInternational:@"MKD failed"];
	}
	else {
		id ds = [m_fileView dataSource];
		const BOOL listResult = [self fullpathListFileDisplay:ds];
		if (listResult) {
			// set cursor.
			for (NSUInteger i = 0; i <[ds displayCount]; i++) {
				id fd = [ds displayObjectAtIndex:i];
				if ([dirName compare:[fd fileName]] == NSOrderedSame) {
					[self setFileCursorPosition:i];
					break;
				}
			}
		}
		else {
			[MyUtil messageBoxInternational:@"LIST failed"];
		}
		// relaod data.
		[m_fileView reloadData];
		[m_directoryView reloadData];
	}

	return ret;
}

/**
 * store file.
 * @param[in] filePath file path to store.
 * @return YES if success, NO if failed.
 */
- (BOOL)storeFile:(NSString *)filePath
{
	// to obtain file path.
	NSString *const fileName = [filePath lastPathComponent];

	id ds = [m_fileView dataSource];
	BOOL isDir = NO;
	if ([ds hasPath:fileName isDir:&isDir] == YES) {
		// same file exist.
		if ([self isSkipRemoteFile:fileName forDirectory:isDir] == YES) {
			return YES;	// skip.
		}
	}

	// update dialog box title.
	[self updateProgressTitle:fileName directory:[m_client currentDirectory] withFormat:@"Storing:"];

	// store file.
	NSData *const fileData = [NSData dataWithContentsOfFile:filePath options:0 error:nil];
	return [m_client stor:fileName content:fileData];
}

/**
 * store directory.
 * @param[in] dirPath directory path to store.
 * @return YES if success, NO if failed.
 */
- (BOOL)storeDirectory:(NSString *)dirPath
{
	// to obtain file path.
	NSString *const fileName = [dirPath lastPathComponent];

	return [self storeDirectoryRecursive:dirPath relPath:fileName];
}

/**
 * store directory recursivelly.
 * @param[in] dirPath directory path to store.
 * @param[in] relPath relative path from operation start point.
 * @return YES if success, NO if failed.
 */
- (BOOL)storeDirectoryRecursive:(NSString *)dirPath relPath:(NSString *)relPath
{
	// at first take care of it self.
	const NSUInteger ret = [self makeDirectory:relPath];
	if (ret == -1) {
		return NO;	// error.
	}
	else if (ret == 0) {
		return YES;	// skip.
	}

	BOOL success = YES;
	NSFileManager *fileManager = [NSFileManager defaultManager];
	NSDirectoryEnumerator *enumerator = [fileManager enumeratorAtPath:dirPath];
	NSString *fullPath = nil;
	NSString *newRelPath = nil;

	// get first file.
	NSString *fileName = [enumerator nextObject];
	while (fileName != nil) {

		// subdirectories will take care recursively.
		if ([fileName rangeOfString:@"/"].length > 0) {
			// get next file.
			fileName = [enumerator nextObject];
			continue;
		}
		else if ([fileName characterAtIndex:0] == '.'){
			// hidden file.
			fileName = [enumerator nextObject];
			continue;
		}


		// create path.
		fullPath = [dirPath stringByAppendingPathComponent:fileName];
		newRelPath = [relPath stringByAppendingPathComponent:fileName];
		//NSLog(@"newRelPath = %@", newRelPath);

		// is directory or physical file.
		const BOOL isDir = [MyUtil isDirectory:fullPath];
		// if directory recursive store.
		if (isDir != NO) {
			// is directory.
			success = [self storeDirectoryRecursive:fullPath relPath:newRelPath];
			if (success == NO) {
				break;
			}
		}
		else {
			// is physical file.
			// get parent directory first.
			id ds = [m_fileView dataSource];
			id fd = [ds fileDisplay:relPath];
			[m_client setCurrentDirectory:[fd fullPath]];
			// goto find same file name of remote path.
			BOOL isDir = NO;
			if ([fd hasPath:fileName isDir:&isDir] == YES) {
				// same file exist.
				if ([self isSkipRemoteFile:fileName forDirectory:isDir] == YES) {

					// get next file.
					fileName = [enumerator nextObject];
					continue;
				}
			}
			// update dialog box title.
			[self updateProgressTitle:fileName directory:[m_client currentDirectory] withFormat:@"Storing:"];

			// store file.
			NSData *const fileData = [NSData dataWithContentsOfFile:fullPath options:0 error:nil];
			success = [m_client stor:fileName content:fileData];
			if (success == NO) {
				[MyUtil messageBoxInternational:@"STOR failed"];
				break;
			}
		}

		// get next file.
		fileName = [enumerator nextObject];
		if ([self isCanceled]) {
			break;	// user abort.
		}
	}

	return success;
}

/**
 * make directory if neccessary.
 * @param[in] dirPath directory path to store.
 * @return 0 if skip, 1 if overwrite, -1 if error.
 */
- (NSUInteger)makeDirectory:(NSString *)dirPath
{
	id ds = [m_fileView dataSource];
	id fd = [ds fileDisplay:dirPath];
	BOOL makeDir = NO;
	if ((fd != nil) && ([fd isDirectory] == NO)) {
		// same name remote file exist.
		if ([self isSkipRemoteFile:dirPath forDirectory:YES] == YES) {
			// user choose skip.
			return 0;
		}
		// remove file.
		NSString *const remotePath = [[ds fullPath] stringByAppendingPathComponent:dirPath];
		//NSLog(@"remotePath = %@", remotePath);
		const NSRange range = [remotePath rangeOfString:@"/" options:NSBackwardsSearch];
		NSString *const currentPath = [remotePath substringToIndex:range.location];
		//NSLog(@"currentPath = %@", currentPath);
		NSString *const rmPath = [remotePath substringFromIndex:range.location + 1];
		//NSLog(@"rmPath = %@", rmPath);
		[m_client setCurrentDirectory:currentPath];

		if ([m_client rm:rmPath] == NO) {
			return -1;
		}
		// file is successfully removed.
		makeDir = YES;
	}
	if (fd == nil) {
		makeDir = YES;
	}

	BOOL success = YES;
	if (makeDir != NO) {
		// create directory.
		NSString *remotePath = [[ds fullPath] stringByAppendingPathComponent:dirPath];
		//NSLog(@"remotePath = %@", remotePath);
		const NSRange range = [remotePath rangeOfString:@"/" options:NSBackwardsSearch];
		NSString *const currentPath = [remotePath substringToIndex:range.location];
		//NSLog(@"currentPath = %@", currentPath);
		NSString *const mkdPath = [remotePath substringFromIndex:range.location + 1];
		//NSLog(@"mkdPath = %@", mkdPath);
		[m_client setCurrentDirectory:currentPath];
		success = [m_client mkd:mkdPath];	// making remote directory.

		// LIST for parent directory as a current directory.
		if (success != NO) {
			const NSRange range = [dirPath rangeOfString:@"/" options:NSBackwardsSearch];
			if (range.length > 0) {
				remotePath = [dirPath substringToIndex:range.location];
				fd = [ds fileDisplay:remotePath];	// LIST parent directory.
				//[self fullpathListFileDisplay:fd];
				[fd addDummyDirectory:mkdPath];
			}
			else {
				// first LIST of MKD.
				[self fullpathListFileDisplay:ds];
			}
		}
	}
	else {
		if ([fd isDirectory] != NO) {
			// if directory, update.
			[self fullpathListFileDisplay:fd];
		}
	}
	if (success == NO) {
		return -1;
	}
	// overwrite.
	return 1;
}

/**
 * get current file name.
 * @return Current selected file name.
 */
- (NSString *)currentFileName
{
	id ds = [m_fileView dataSource];
	const NSInteger sel = [m_fileView selectedRow];
	id fd = [ds displayObjectAtIndex:sel];
	return [fd fileTitle];
}

/**
 * get current file mode.
 * @return Current selected file mode.
 */
- (NSString *)currentFileMode
{
	id ds = [m_fileView dataSource];
	const NSInteger sel = [m_fileView selectedRow];
	id fd = [ds displayObjectAtIndex:sel];
	return [NSString stringWithFormat:@"%03d", [fd fileMode]];
}

/**
 * get file slection count.
 * @return selected file count.
 */
- (NSUInteger)fileSelectionCount
{
	NSIndexSet *const indexSet = [m_fileView selectedRowIndexes];
	return [indexSet count];
}

/**
 * is connected to server or not.
 * @return YES if connected to server, NO if not.
 */
- (BOOL)isConnected
{
	return [m_client isConnect];
}

/**
 * set file cursor position.
 * @param[in] pos file view cursor pos.
 */
- (void)setFileCursorPosition:(NSUInteger)pos
{
	NSIndexSet *const indexSet = [NSIndexSet indexSetWithIndex:pos];
	[m_fileView selectRowIndexes:indexSet byExtendingSelection:NO];
}

/**
 * is user choose to skip or overwrite?
 * @param[in] path asking path name to overwrite.
 * @param[in] dir YES if the operation is for directory, NO if not.
 * @return YES for skip, NO for overwrite.
 */
- (BOOL)isSkipFile:(NSString *)path forDirectory:(BOOL)dir
{
	NSFileManager *const fileManager = [NSFileManager defaultManager];
	BOOL isDir = NO;
	const BOOL exist = [fileManager fileExistsAtPath:path isDirectory:&isDir];
	if (exist == NO) {
		// path not found, so you cant write it.
		return NO;
	}
	if (m_ask == NO) {
		return m_skip;
	}
	if (dir != NO) {
		if (isDir != NO) {
			return NO;	// overwrite.
		}
	}
	// ask to user.
	[m_askSkipOverwriteDlgController setFilePath:path];
	[m_askSkipOverwriteDlgController showModal];
	return m_skip;
}

/**
 * is user choose to skip or overwrite? remote version.
 * @param[in] path asking path name to overwrite.
 * @param[in] dir YES if the operation is for directory, NO if not.
 * @return YES for skip, NO for overwrite.
 */
- (BOOL)isSkipRemoteFile:(NSString *)path forDirectory:(BOOL)dir
{
	if (m_ask == NO) {
		return m_skip;
	}
	// ask to user.
	[m_askSkipOverwriteDlgController setFilePath:path];
	[m_askSkipOverwriteDlgController showModal];
	return m_skip;
}

/**
 * users reply of skip or overwrite.
 * @param[in] skip if YES skip, if NO overwrite.
 * @param[in] doSame if YES do not ask any more, if NO ask if same situation occurred.
 */
- (void)replaySkip:(BOOL)skip doSame:(BOOL)doSame
{
	m_skip = skip;
	m_ask = (doSame == NO) ? YES: NO;
}

/**
 * fullpath list of file display.
 * @param[in] fileDisplay FileDisplayCollection object.
 * @return YES if success, NO if failed.
 */
- (BOOL)fullpathListFileDisplay:(id)fileDisplay
{
	[fileDisplay invalidate];
	const BOOL ret = [m_client list:[fileDisplay fullPath]];
	if (ret) {
		[fileDisplay updateAll];
	}
	return ret;
}

/**
 * update directory view.
 */
- (void)updateDirectoryView
{
	[m_directoryView reloadData];
}

/**
 * LIST directory.
 * @param[in] dir directory name.
 * @param[in] file file name.
 * @return FileDisplayObject.
 */
- (id)listDir:(NSString *)dir file:(NSString *)file
{
	id dirDisp = [self findDirectory:dir];	// find directory.
	if ((dirDisp == nil) || ([dirDisp isDirectory] == NO)) {
		return nil;
	}
	// LIST.
	NSIndexSet *const indexSet = [m_fileView selectedRowIndexes];
	[self directorySelectionHasChanged:dirDisp];
	[m_fileView selectRowIndexes:indexSet byExtendingSelection:NO];
	return [dirDisp fileDisplay:file];	// find file.
}

/**
 * find directory.
 * @param[in] dir directory name.
 * @return FileDisplayObject.
 */
- (id)findDirectory:(NSString *)dir
{
	return [m_fileDisplayController retrieveFileDisplay:dir];	// find directory.
}

/**
 * store file with content.
 * @param[in] filePath file name.
 * @param[in] dir source directory name.
 * @param[in] content content of the file.
 * @return YES if success, NO if failed.
 */
- (BOOL)storeFile:(NSString *)filePath directory:(NSString *)dir content:(NSData *)content
{
	FileDisplayCollection *const dirDisp = [self findDirectory:m_savedCurDir];
	if ((dirDisp == nil) || ([dirDisp isDirectory] == NO)) {
		return NO;
	}
	// find file.
	FileDisplay *const file = [dirDisp fileDisplay:filePath];
	if (file != nil) {
		if ([self isSkipRemoteFile:filePath forDirectory:NO] == YES) {
			return YES;	// skip.
		}
	}

	// update dialog box title.
	[self updateProgressTitle:filePath directory:[m_client currentDirectory] withFormat:@"Storing:"];

	// setup client.
	[m_client setCurrentDirectory:m_savedCurDir];
	return [m_client stor:filePath content:content];
}

/**
 * copy directory.
 * @param[in] fileDisplay FileDisplayCollection class.
 * @param[in] srcController source controllers.
 * @return YES if success, NO if failed.
 */
- (BOOL)copyDirectory:(id)fileDisplay srcController:(id)srcController
{
	BOOL ret = NO;
	NSString *const fullPath = [fileDisplay fullPath];
	@try {
		[fullPath retain];
		[srcController silentDirectorySelectionHasChanged:fileDisplay];
		fileDisplay = [srcController findDirectory:fullPath];

		// mkdir
		NSString *const myPath = [m_savedCurDir stringByAppendingPathComponent:[fileDisplay fileName]];
		FileDisplayCollection *myDir = [self findDirectory:myPath];
		if (myDir != nil) {
			if ([self isSkipRemoteFile:[myDir fileName] forDirectory:YES] == YES) {
				return YES;	// skip.
			}
			if ([myDir isDirectory] == NO) {
				// remove file.
				[m_client setCurrentDirectory:m_savedCurDir];
				if ([m_client rm:[myDir fileName]] == NO) {
					NSString *const msg = [NSString stringWithFormat:NSLocalizedString(@"remove failed", @"remove failed"), [myDir fileName]];
					@throw [NSException exceptionWithName:@"recursiveCopyDirectory" reason:msg userInfo:nil];
				}
				myDir = nil;
			}
		}
		if (myDir == nil) {
			[m_client setCurrentDirectory:m_savedCurDir];
			const BOOL mkDirResult = [m_client mkd:[fileDisplay fileName]];
			if (mkDirResult == NO) {
				@throw [NSException exceptionWithName:@"copyDirectory" reason:@"MKD failed" userInfo:nil];
			}
		}
		// refresh data.
		id fd = [self findDirectory:m_savedCurDir];
		[self fullpathListFileDisplay:fd];
		// retrieve directory.
		myDir = [self findDirectory:myPath];
		if (myDir == nil) {
			@throw [NSException exceptionWithName:@"copyDirectory" reason:@"LIST failed" userInfo:nil];
		}
		
		// recursive copy.
		if ([self recursiveCopyDirectory:myDir relativePath:[myDir fileName] srcDisplay:fileDisplay srcController:srcController] == NO) {
			@throw [NSException exceptionWithName:@"copyDirectory" reason:@"unknown error" userInfo:nil];
		}

		ret = YES;
	}
	@catch (NSException *e) {
		@throw e;
	}
	@finally {
		[fullPath release];
	}

	return ret;
}

/**
 * recursive copy.
 * @param[in] myDisplay this instant's (copy destincation) FileDisplay object.
 * @param[in] relativePath relative path from copy root.
 * @param[in] srcDisplay copy source instant's FileDisplay object.
 * @param[in] srcController copy source controller object.
 * @return YES if success, NO if failed.
 */
- (BOOL)recursiveCopyDirectory:(id)myDisplay relativePath:(NSString *)relativePath srcDisplay:(id)srcDisplay srcController:(id)srcController
{
	for (NSUInteger i = 0; i < [srcDisplay displayCount]; i++) {
		id fd = [srcDisplay displayObjectAtIndex:i];
		[self fullpathListFileDisplay:myDisplay];
		id myFd = [myDisplay fileDisplay:[fd fileTitle]];
		if ([fd isDirectory] != NO) {
			// source is directory.
			NSString *const path = [relativePath stringByAppendingPathComponent:[fd fileTitle]];
			if (myFd != nil) {
				if ([self isSkipRemoteFile:path forDirectory:YES] == YES) {
					continue;	// user choose skip.
				}
				if ([myFd isDirectory] == NO) {
					// remove physical file.
					[m_client setCurrentDirectory:[myDisplay fullPath]];
					if ([m_client rm:[myFd fileName]] == NO) {	// RM
						NSString *const msg = [NSString stringWithFormat:NSLocalizedString(@"remove failed", @"remove failed"), [fd fileName]];
						@throw [NSException exceptionWithName:@"recursiveCopyDirectory" reason:msg userInfo:nil];
					}
					myFd = nil;	// goto LIST.
				}
			}
			if (myFd == nil) {
				NSString *const fullPath = [myDisplay fullPath];
				[m_client setCurrentDirectory:fullPath];
				if ([m_client mkd:[fd fileTitle]] == NO) {
					@throw [NSException exceptionWithName:@"recursiveCopyDirectory" reason:@"MKD failed" userInfo:nil];
				}
				// LIST.
				[self fullpathListFileDisplay:myDisplay];
				// myFd again.
				myFd = [myDisplay fileDisplay:[fd fileName]];
				if (myFd == nil) {
					@throw [NSException exceptionWithName:@"recursiveCopyDirectory" reason:@"MKD failed" userInfo:nil];
				}
			}

			// recursive call.
			[srcController fullpathListFileDisplay:fd];
			if ([self recursiveCopyDirectory:myFd relativePath:path srcDisplay:fd srcController:srcController] == NO) {
				@throw [NSException exceptionWithName:@"recursiveCopyDirectory" reason:@"MKD failed" userInfo:nil];
			}
		}
		else {
			// source is physical file.
			if (myFd != nil) {
				NSString *const path = [NSString stringWithFormat:@"%@/%@", relativePath, [fd fileName]];
				if ([self isSkipRemoteFile:path forDirectory:NO] == YES) {
					continue;	// user choose skip.
				}
				if ([myFd isDirectory] != NO) {
					// remove directory.
					[m_client setCurrentDirectory:[myDisplay fullPath]];
					if ([m_client rmd:[myFd fileName]] == NO) {	// RMD
						NSString *const msg = [NSString stringWithFormat:NSLocalizedString(@"remove failed", @"remove failed"), [fd fileName]];
						@throw [NSException exceptionWithName:@"recursiveCopyDirectory" reason:msg userInfo:nil];
					}
				}
			}
			// RETR and STOR.
			NSMutableData *content = nil;
			@try {
				// update dialog box title.
				[self updateProgressTitle:[fd fileTitle] directory:[m_client currentDirectory] withFormat:@"Fetching:"];

				// RETR.
				//NSString *const srcPath = [FileDisplayCollection buildFullPathDir:[srcDisplay fullPath] file:[fd fileName]];
				//NSLog(@"srcPath = %@", srcPath);
				id srcClient = [srcController client];
				[srcClient setCurrentDirectory:[srcDisplay fullPath]];
				content = [[NSMutableData alloc] init];	// allocate data.
				if ([srcClient retr:[fd fileTitle] content:content] == NO) {
					@throw [NSException exceptionWithName:@"recursiveCopyDirectory" reason:@"RETR failed" userInfo:nil];
				}

				// update dialog box title.
				[self updateProgressTitle:[fd fileTitle] directory:[m_client currentDirectory] withFormat:@"Storing:"];

				// STOR.
				id myClient = [self client];
				[myClient setCurrentDirectory:[myDisplay fullPath]];
				if ([myClient stor:[fd fileTitle] content:content] == NO) {
					@throw [NSException exceptionWithName:@"recursiveCopyDirectory" reason:@"STOR failed" userInfo:nil];
				}
			}
			@catch (NSException *e) {
				@throw e;
			}
			@finally {
				[content release];
			}
		}
		if ([self isCanceled]) {
			break;	// user abort.
		}
	}
	return YES;
}

/**
 * long process strat.
 */
- (void)longProcessStart
{
	m_ask = YES;
	m_skip = YES;
	m_cancel = NO;
	[m_controlDisabler disalbeButton:m_connectButton];
	[m_controlDisabler disableWindowCloseStyle:m_mainWindow];

	[self setCanClose:NO];
}

/**
 * disconnected.
 */
- (void)disconnected
{
	[m_fileDisplayController setCurrentDirectory:@"/"];
	[m_client connect:NO];
	// disconnect.
	[m_connectButton setTitle:NSLocalizedString(@"Connect", @"Connect")];
	[m_settingButton setEnabled:YES];
	// remove view items.
	[m_fileDisplayController removeAllObjects];
	[m_fileView setDataSource:nil];
	[m_fileView setDelegate:nil];
	// delegate.
	id appDelegate = [NSApp delegate];
	[appDelegate controllerDisconnect:self];

	// update display.
	[m_fileView reloadData];
	[self resetDirectoryView];
}

/**
 * selection double clicked.
 * @param[in] ds newly selected file display object.
 */
- (void)directoryDoubleClicked:(id)fd
{
	//NSLog(@"directoryDoubleClicked:%@", [fd fullPath]);
	[self setCurrentDirectory:[fd fullPath]];
	[self directorySelectionHasChanged:fd];
	[m_fileDisplayController setCurrentDirectory:[fd fullPath]];
	[m_fileDisplayController expandOutlineView:m_directoryView];

	// find a leaf of new selected item from m_directoryView.
	NSIndexSet *const indexSet = [m_directoryView selectedRowIndexes];
	const NSUInteger index = [indexSet firstIndex];
	[m_directoryView scrollRowToVisible:index];	// make new selection visible.
}

/**
 * select added file.
 * @param[in] addedFiles added files.
 */
- (void)selectAddedFiles:(NSArray *)addedFiles
{
	id ds = [m_fileView dataSource];
	NSMutableIndexSet *const indexSet = [NSMutableIndexSet indexSet];
	for (NSUInteger i = 0; i < [addedFiles count]; i++) {
		NSString *const filePath = [[addedFiles objectAtIndex:i] lastPathComponent];
		for (NSUInteger j = 0; j < [ds displayCount]; j++) {
			id fd = [ds displayObjectAtIndex:j];
			if ([filePath compare:[fd fileName]] == NSOrderedSame) {
				[indexSet addIndex:j];
				break;
			}
		}
	}
	[m_fileView selectRowIndexes:indexSet byExtendingSelection:NO];
	// at least, make the first selected row visible.
	[m_fileView scrollRowToVisible:[indexSet firstIndex]];
}

/**
 * update progress title.
 * @param[in] file file name.
 * @param[in] dir directory name.
 * @param[in] format a format string.
 */
- (void)updateProgressTitle:(NSString *)file directory:(NSString *)dir withFormat:(NSString *)format
{
	if (file) {
		NSString *const title = [NSString stringWithFormat:NSLocalizedString(format, format), file, dir];
		[self setProgressTitle:title];
	}
	else {
		[self setProgressTitle:NSLocalizedString(format, format)];
	}
	[self performSelectorOnMainThread:@selector(mainThreadUpdateProgressTitle:) withObject:m_progressTitle waitUntilDone:NO];
}

/**
 * update progress title with main thread.
 * @param[in] title a progress title.
 */
- (void)mainThreadUpdateProgressTitle:(NSString *)title;
{
	[m_progress setStringValue:[self progressTitle]];
}

/**
 * end sub process.
 */
- (void)endSubProcess
{
	[self performSelectorOnMainThread:@selector(mainThreadEndSubProcess:) withObject:nil waitUntilDone:NO];
}

/**
 * end sub process with main thread.
 * @param[in] some nil
 */
- (void)mainThreadEndSubProcess:(id)some
{
	[m_progress setStringValue:@""];
	[m_controlDisabler resetAll];
	[m_fileView reloadData];
	[m_directoryView reloadData];

	[self setCanClose:YES];
}

/**
 * set can close.
 * @param[in] canClose YES can close, NO cannot close.
 */
- (void)setCanClose:(BOOL)canClose
{
	[m_cancelButton setEnabled:canClose == NO ? YES : NO];
	[m_controlDisabler setCanClose:canClose];
	[m_fileView setEnabled:canClose];
	[m_directoryView setEnabled:canClose];

}

// awake from XIB.
- (void)awakeFromNib
{
	[m_owner setController:self];

	// resize split window.
	[m_splitView setPosition:200.0 ofDividerAtIndex:0];
	[m_directoryView setDataSource:m_fileDisplayController];
	[m_directoryView setDelegate:m_fileDisplayController];
	// resize file view.
	NSTableColumn *column = nil;
	column = [m_fileView tableColumnWithIdentifier:@"Name"];
	[column setWidth:200.0];
	column = [m_fileView tableColumnWithIdentifier:@"Owner"];
	[column setWidth:80.0];
	column = [m_fileView tableColumnWithIdentifier:@"Group"];
	[column setWidth:60.0];
	column = [m_fileView tableColumnWithIdentifier:@"Date"];
	[column setWidth:80.0];
	column = [m_fileView tableColumnWithIdentifier:@"Size"];
	[column setWidth:60.0];
	column = [m_fileView tableColumnWithIdentifier:@"Mode"];
	[column setWidth:60.0];

	[m_renameDialogController setProcess:RENAME_DIALOG_UNKNOWN];
	[m_statusField setStringValue:@""];
	[m_progress setStringValue:@""];
}

// constructor.
- (id)init
{
	self = [super init];
	if (self) {
		m_client = [[ShsslClient alloc] init];
		[m_client setController:self];
		m_fileDisplayController = [[FileDisplayController alloc] init];
		[m_fileDisplayController setController:self];
		m_savedCurDir = @"";
		[m_savedCurDir retain];
		[m_progress setStringValue:@""];
		m_cancel = NO;
		m_controlDisabler = [[ControlDisabler alloc] init];
	}
	return self;
}

// destructor.
- (void)dealloc
{
	if ([m_client isConnect] != NO) {
		[self connectButton:nil];
	}
	[m_controlDisabler release];
	[m_progressThread release];
	[m_savedCurDir release];
	[m_fileDisplayController release];
	[m_client release];
	[super dealloc];
}

@end
