/**@file
 * XmlDelegater.m
 * @brief Evend driven XML parser delegation class.
 * @author toru kageyama
 * Created by Toru Kageyama on 10/09/07.
 * Copyright 2010 Comona Co., Ltd. All rights reserved.
 */

#import "XmlDelegater.h"
#import "MyDocument.h"

@implementation XmlDelegater

/**
 * set event target.
 * @param[in] target event target instance.
 */
- (void)setTarget:(id)target
{
	m_target = target;
}

// XML handler methods.

/**
 * XML element started.
 * @param[in] parser parser itself.
 * @param[in] elementName XML element name.
 * @param[in] namespaceURI URI of namespace, if assigned.
 * @param[in] qualifiedName qualified name.
 * @param[in] attributeDict XML attributes.
 */
- (void)parser:(NSXMLParser *)parser didStartElement:(NSString *)elementName namespaceURI:(NSString *)namespaceURI
				qualifiedName:(NSString *)qualifiedName attributes:(NSDictionary *)attributeDict
{
	//NSLog(@"*** Start Element(%@)", elementName);
	[m_stack addObject:elementName];
	// call document.
	[m_target addElement:m_stack withDictionary:attributeDict];
}

/**
 * XML element ended.
 * @param[in] parser parser itself.
 * @param[in] elementName XML element name.
 * @param[in] namespaceURI URI of namespace, if assigned.
 * @param[in] qualifiedName qualified name.
 */
- (void)parser:(NSXMLParser *)parser didEndElement:(NSString *)elementName namespaceURI:(NSString *)namespaceURI
				qualifiedName:(NSString *)qualifiedName
{
	//NSLog(@"*** End Element(%@)", elementName);
	[m_stack removeObjectAtIndex:[m_stack count] - 1];
}

/**
 * XML item name inside element.
 * @param[in] parser parser itself.
 * @param[in] string element inside string.
 */
- (void)parser:(NSXMLParser *)parser foundCharacters:(NSString *)string
{
}

// start document.
- (void)parserDidStartDocument:(NSXMLParser *)parser
{
	//NSLog(@"*** Start Document ***");
}

// end document.
- (void)parserDidEndDocument:(NSXMLParser *)parser
{
	//NSLog(@"*** End Document ***");
}

// constructor.
- (id)init
{
	self = [super init];
	if (self) {
		m_stack = [[NSMutableArray array] retain];
	}
	return self;
}

// Destructor.
- (void)dealloc
{
	[m_stack release];
	[super dealloc];
}

@end
